from copy import copy, deepcopy
from datetime import date as Date, datetime as Datetime, timedelta as Timedelta
from decimal import Decimal
from io import TextIOWrapper
from zipfile import ZIP_DEFLATED, ZIP_STORED, ZipFile

from odio import OdioException
from odio.v1_3.mapping import MAP
from odio.v1_3.xml import Element, GeneralElement, document_to_str, write_document

ELEMENT_MAP = {}


class Formula:
    def __init__(self, formula):
        self.formula = formula

    def __repr__(self):
        return f"odio.Formula('{self.formula}')"

    def __str__(self):
        return self.formula

    def __eq__(self, other):
        return isinstance(other, Formula) and self.formula == other.formula


class BaseElement(Element):
    def __init__(self, *children, attributes=None):
        allowed_attributes = ELEMENT_MAP[self.tag_name]["allowed_attributes"]
        allowed_children = ELEMENT_MAP[self.tag_name]["allowed_children"]
        super().__init__(
            *children,
            attributes=attributes,
            allowed_attributes=allowed_attributes,
            allowed_children=allowed_children,
        )


class LineBreak(BaseElement):
    def __init__(self, fill=True):
        super().__init__("text:line-break")


class Meta(BaseElement):
    def __init__(self, *children, attributes=None, fill=True):
        super().__init__("text:meta", *children, attributes=attributes)


class A(BaseElement):
    tag_name = "text:a"

    def __init__(self, *children, href=None, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            if href is not None:
                self.attributes["xlink:href"] = href
                self.attributes["xlink:type"] = "simple"

    @property
    def href(self):
        return self.attributes.get("xlink:href")


class H1(BaseElement):
    tag_name = "text:h"

    def __init__(self, *children, attributes=None, fill=True, style_name=None):
        super().__init__(*children, attributes=attributes)
        if fill:
            if style_name is not None:
                self.attributes["text:style-name"] = style_name
            self.attributes["text:outline-level"] = "1"

    @property
    def level(self):
        return self.attributes.get("text:outline-level")


class H2(BaseElement):
    tag_name = "text:h"

    def __init__(self, *children, attributes=None, fill=True, style_name=None):
        super().__init__(*children, attributes=attributes)
        if fill:
            if style_name is not None:
                self.attributes["text:style-name"] = style_name
            self.attributes["text:outline-level"] = "2"

    @property
    def level(self):
        return self.attributes.get("text:outline-level")


class H3(BaseElement):
    tag_name = "text:h"

    def __init__(self, *children, attributes=None, fill=True, style_name=None):
        super().__init__(*children, attributes=attributes)
        if fill:
            if style_name is not None:
                self.attributes["text:style-name"] = style_name
            self.attributes["text:outline-level"] = "3"

    @property
    def level(self):
        return self.attributes.get("text:outline-level")


class H4(BaseElement):
    tag_name = "text:h"

    def __init__(self, *children, attributes=None, fill=True, style_name=None):
        super().__init__(*children, attributes=attributes)
        if fill:
            if style_name is not None:
                self.attributes["text:style-name"] = style_name
            self.attributes["text:outline-level"] = "4"

    @property
    def level(self):
        return self.attributes.get("text:outline-level")


class H5(BaseElement):
    tag_name = "text:h"

    def __init__(self, *children, attributes=None, fill=True, style_name=None):
        super().__init__(*children, attributes=attributes)
        if fill:
            if style_name is not None:
                self.attributes["text:style-name"] = style_name
            self.attributes["text:outline-level"] = "5"

    @property
    def level(self):
        return self.attributes.get("text:outline-level")


class H6(BaseElement):
    tag_name = "text:h"

    def __init__(self, *children, attributes=None, fill=True, style_name=None):
        super().__init__(*children, attributes=attributes)
        if fill:
            if style_name is not None:
                self.attributes["text:style-name"] = style_name
            self.attributes["text:outline-level"] = "6"

    @property
    def level(self):
        return self.attributes.get("text:outline-level")


class H(BaseElement):
    LEVELS = {
        "1": H1,
        "2": H2,
        "3": H3,
        "4": H4,
        "5": H5,
        "6": H6,
    }
    tag_name = "text:h"

    def __init__(
        self, *children, level=None, attributes=None, fill=True, style_name=None
    ):
        super().__init__(*children, attributes=attributes)
        if fill:
            if style_name is not None:
                self.attributes["text:style-name"] = style_name
            if level is not None:
                self.attributes["text:outline-level"] = level

    @property
    def level(self):
        return self.attributes.get("text:outline-level")

    def unpack(self):
        level = self.level
        if level is None:
            return self
        else:
            cls = H.LEVELS[level]
            return cls(*self.children, attributes=self.attributes, fill=False)


class P(BaseElement):
    tag_name = "text:p"

    def __init__(self, *children, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)


class Manifest(BaseElement):
    tag_name = "manifest:manifest"

    def __init__(self, *children, media_type=None, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["manifest:version"] = "1.3"
            self.attributes["xmlns:manifest"] = (
                "urn:oasis:names:tc:opendocument:xmlns:manifest:1.0"
            )
            self.append_entry("/", media_type)

    def append_entry(self, full_path, media_type):
        self.children.append(
            ManifestFileEntry(full_path=full_path, media_type=media_type)
        )


class ManifestFileEntry(BaseElement):
    tag_name = "manifest:file-entry"

    def __init__(
        self, *children, attributes=None, full_path=None, media_type=None, fill=True
    ):
        super().__init__(attributes=attributes, *children)
        if fill:
            if full_path is not None:
                self.attributes["manifest:full-path"] = full_path
            if media_type is not None:
                self.attributes["manifest:media-type"] = media_type

    @property
    def full_path(self):
        return self.attributes["manifest:full-path"]

    @property
    def media_type(self):
        return self.attributes["manifest:media-type"]


class Cell(BaseElement):
    tag_name = "table:table-cell"

    def __init__(
        self,
        *children,
        value=None,
        currency=None,
        is_percentage=False,
        attributes=None,
        fill=True,
    ):
        super().__init__(attributes=attributes, *children)
        if fill:
            if value is not None:
                if isinstance(value, bool):
                    self.attributes["office:value-type"] = "boolean"
                    self.attributes["office:boolean-value"] = (
                        "true" if value else "false"
                    )

                elif isinstance(value, (Decimal, float, int)):
                    if currency is not None:
                        self.attributes["office:currency"] = currency
                        self.attributes["office:value-type"] = "currency"
                    elif is_percentage:
                        self.attributes["office:value-type"] = "percentage"
                    else:
                        self.attributes["office:value-type"] = "float"

                    self.attributes["office:value"] = str(value)

                elif isinstance(value, Datetime):
                    self.attributes["table:style-name"] = "cell_date"
                    self.attributes["office:value-type"] = "date"
                    self.attributes["office:date-value"] = value.strftime(
                        "%Y-%m-%dT%H:%M:%S"
                    )

                elif isinstance(value, Date):
                    self.attributes["table:style-name"] = "cell_date"
                    self.attributes["office:value-type"] = "date"
                    self.attributes["office:date-value"] = value.strftime("%Y-%m-%d")

                elif isinstance(value, Timedelta):
                    self.attributes["table:style-name"] = "cell_date"
                    self.attributes["office:value-type"] = "time"
                    self.attributes["office:time-value"] = value.strftime("%H:%M:%S")

                elif value is None:
                    self.attributes["office:value-type"] = "void"

                elif isinstance(value, str):
                    self.attributes["office:value-type"] = "string"
                    self.attributes["office:string-value"] = value

                elif isinstance(value, Formula):
                    self.attributes["table:formula"] = f"of:{value}"

    @property
    def value(self):
        if "office:value-type" in self.attributes:
            if self.value_type == "boolean":
                return self.attributes["office:boolean-value"] == "true"

            elif self.value_type == "currency":
                return Decimal(self.attributes["office:value"])

            elif self.value_type == "date":
                val = self.attributes["office:date-value"]
                if len(val) == 10:
                    return Date.fromisoformat(val)
                else:
                    return Datetime.fromisoformat(val)

            elif self.value_type == "float":
                val = self.attributes["office:value"]
                return float(val)

            elif self.value_type == "percentage":
                val = self.attributes["office:value"]
                return float(val)

            elif self.value_type == "string":
                return (
                    self.attributes["office:string-value"]
                    if "office:string-value" in self.attributes
                    else ""
                ) + self.to_text()

            elif self.value_type == "time":
                return val.strptime(self.attributes["office:time-value"], "%H:%M:%S")

            elif self.value_type == "void":
                return None

            else:
                return None

        elif "table:formula" in self.attributes:
            return Formula(self.attributes["table:formula"][3:])
        else:
            return None

    @property
    def value_str(self):
        if "office:value-type" in self.attributes:
            if self.value_type == "boolean":
                return self.attributes["office:boolean-value"]

            elif self.value_type in ("currency", "date", "float", "percentage"):
                return self.attributes["office:value"]

            elif self.value_type == "string":
                return (
                    self.attributes["office:string-value"]
                    if "office:string-value" in self.attributes
                    else ""
                ) + self.to_text()

            elif self.value_type == "time":
                return self.attributes["office:time-value"]

            elif self.value_type == "void":
                return None

            else:
                return None

        elif "table:formula" in self.attributes:
            return self.attributes["table:formula"][3:]
        else:
            return None

    @property
    def currency(self):
        return self.attributes.get("office:currency")

    @property
    def value_type(self):
        return self.attributes.get("office:value-type")

    @property
    def formula(self):
        try:
            f = self.attributes["table:formula"]
            eq_idx = f.index("=")
            return f[eq_idx:]
        except KeyError:
            return None


NAMESPACES = {
    "urn:oasis:names:tc:opendocument:xmlns:animation:1.0": "anim",
    "urn:oasis:names:tc:opendocument:xmlns:chart:1.0": "chart",
    "urn:oasis:names:tc:opendocument:xmlns:config:1.0": "config",
    "urn:oasis:names:tc:opendocument:xmlns:database:1.0": "db",
    "urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0": "dr3d",
    "urn:oasis:names:tc:opendocument:xmlns:drawing:1.0": "draw",
    "urn:oasis:names:tc:opendocument:xmlns:form:1.0": "form",
    "urn:oasis:names:tc:opendocument:xmlns:manifest:1.0": "manifest",
    "urn:oasis:names:tc:opendocument:xmlns:meta:1.0": "meta",
    "urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0": "number",
    "urn:oasis:names:tc:opendocument:xmlns:office:1.0": "office",
    "urn:oasis:names:tc:opendocument:xmlns:presentation:1.0": "presentation",
    "urn:oasis:names:tc:opendocument:xmlns:script:1.0": "script",
    "urn:oasis:names:tc:opendocument:xmlns:table:1.0": "table",
    "urn:oasis:names:tc:opendocument:xmlns:text:1.0": "text",
    "urn:oasis:names:tc:opendocument:xmlns:style:1.0": "style",
    "http://docs.oasis-open.org/ns/office/1.3/meta/odf#": "odf",
    "urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0": "fo",
    "urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0": "svg",
    "urn:oasis:names:tc:opendocument:xmlns:smil-compatible:1.0": "smil",
    "http://purl.org/dc/elements/1.1/": "dc",
    "http://www.w3.org/1998/Math/MathML": "math",
    "http://www.w3.org/2002/xforms": "xforms",
    "http://www.w3.org/1999/xlink": "xlink",
    "http://www.w3.org/1999/xhtml": "xhtml",
    "http://www.w3.org/2003/g/data-view#": "grddl",
    "http://docs.oasis-open.org/ns/office/1.3/meta/pkg#": "pkg",
    "urn:oasis:names:tc:opendocument:xmlns:of:1.3": "of",
}


class DocumentMeta(BaseElement):
    tag_name = "office:document-meta"

    def __init__(self, *children, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["office:version"] = "1.3"
            for k, v in NAMESPACES.items():
                if v in {"office", "xlink", "dc", "meta", "of"}:
                    self.attributes[f"xmlns:{v}"] = k
            self.children.append(
                GeneralElement("office:meta", GeneralElement("meta:generator", "Odio"))
            )


class DocumentSettings(BaseElement):
    tag_name = "office:document-settings"

    def __init__(self, *children, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["office:version"] = "1.3"
            for k, v in NAMESPACES.items():
                if v in {"office", "xlink", "config", "of"}:
                    self.attributes[f"xmlns:{v}"] = k


class DocumentStyles(BaseElement):
    tag_name = "office:document-styles"

    def __init__(self, *children, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["office:version"] = "1.3"
            for k, v in NAMESPACES.items():
                if v in {
                    "anim",
                    "chart",
                    "config",
                    "db",
                    "dr3d",
                    "draw",
                    "form",
                    "meta",
                    "number",
                    "office",
                    "presentation",
                    "script",
                    "table",
                    "text",
                    "style",
                    "odf",
                    "fo",
                    "svg",
                    "smil",
                    "dc",
                    "math",
                    "xforms",
                    "xlink",
                    "xhtml",
                    "grddl",
                    "pkg",
                    "of",
                }:
                    self.attributes[f"xmlns:{v}"] = k


class DocumentContent(BaseElement):
    tag_name = "office:document-content"

    def __init__(self, *children, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["office:version"] = "1.3"
            for k, v in NAMESPACES.items():
                if v in {
                    "anim",
                    "chart",
                    "config",
                    "db",
                    "dr3d",
                    "draw",
                    "form",
                    "meta",
                    "number",
                    "office",
                    "presentation",
                    "script",
                    "table",
                    "text",
                    "style",
                    "fo",
                    "svg",
                    "smil",
                    "dc",
                    "math",
                    "xforms",
                    "xlink",
                    "xhtml",
                    "grddl",
                    "of",
                }:
                    self.attributes[f"xmlns:{v}"] = k

    @property
    def body_root(self):
        office_body = self.find_children_by_name("office:body")[0]
        return office_body.children[0]


class Document:
    MIME_TYPES = {
        "office:text": "application/vnd.oasis.opendocument.text",
        "office:spreadsheet": "application/vnd.oasis.opendocument.spreadsheet",
    }

    def __init__(self, content=None, settings=None, meta=None, styles=None):
        self.document_content = DocumentContent() if content is None else content
        self.document_settings = DocumentSettings() if settings is None else settings
        self.document_meta = DocumentMeta() if meta is None else meta
        self.document_styles = DocumentStyles() if styles is None else styles

    def save(self, f, compressed=True):
        compression = ZIP_DEFLATED if compressed else ZIP_STORED
        body_root = self.document_content.body_root
        with ZipFile(f, "w", compression) as z:
            media_type = Document.MIME_TYPES[body_root.tag_name]
            z.writestr("mimetype", media_type, compress_type=ZIP_STORED)

            manifest = Manifest(media_type=media_type)

            for path, media_type, element in (
                ("settings.xml", "text/xml", self.document_settings),
                ("content.xml", "text/xml", self.document_content),
                ("meta.xml", "text/xml", self.document_meta),
                ("styles.xml", "text/xml", self.document_styles),
            ):
                manifest.append_entry(path, media_type)
                with z.open(path, "w") as f:
                    write_document(
                        TextIOWrapper(f, encoding="utf8", newline=""), element
                    )

            z.writestr("META-INF/manifest.xml", document_to_str(manifest))


class Span(BaseElement):
    tag_name = "text:span"

    def __init__(self, *children, attributes=None, fill=True, style_name=None):
        super().__init__(*children, attributes=attributes)
        if fill:
            if style_name is not None:
                self.attributes["text:style-name"] = style_name


class Spreadsheet(BaseElement):
    tag_name = "office:spreadsheet"

    def __init__(self, *children, doc=None, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.doc = doc

    def append_table(self, name, values=None):
        table = Table(name=name, values=values)
        self.children.append(table)
        return table

    def save(self, f, compressed=True):
        return self.doc.save(f, compressed=compressed)

    @property
    def tables(self):
        return self.find_children_by_name("table:table")


class Text(BaseElement):
    tag_name = "office:text"

    def __init__(self, *children, doc=None, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.doc = doc

    def append_table(self, name):
        table = Table(name=name)
        self.children.append(table)
        return table

    def save(self, f, compressed=True):
        return self.doc.save(f, compressed=compressed)

    @property
    def tables(self):
        return self.find_children_by_name("table:table")


class Row(BaseElement):
    tag_name = "table:table-row"

    def __init__(self, *children, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)

    def append_cell(self, value=None, currency=None, is_percentage=False):
        if isinstance(value, Cell):
            cell = value

        else:
            cell = Cell(value=value, currency=currency, is_percentage=is_percentage)

        self.children.append(cell)

    @property
    def cells(self):
        return self.find_children_by_name("table:table-cell")

    @property
    def values(self):
        return [cell.value for cell in self.cells]

    def pack(self):
        new_row = Row(attributes=self.attributes)
        cells = new_row.children

        prev_c = None
        for c in self.children:
            if c == prev_c:
                try:
                    repeats = int(cells[-1].attributes["table:number-columns-repeated"])
                except KeyError:
                    repeats = 1

                cells[-1].attributes["table:number-columns-repeated"] = str(repeats + 1)

            else:
                cells.append(c)

            prev_c = c

        return new_row

    def unpack(self):
        new_row = Row(attributes=self.attributes)
        cells = new_row.children

        for c in self.children:
            try:
                repeats = int(c.attributes["table:number-columns-repeated"])
            except KeyError:
                repeats = 1

            for _ in range(repeats):
                cells.append(copy(c))

        return new_row


class Table(BaseElement):
    tag_name = "table:table"

    def __init__(self, *children, name=None, attributes=None, fill=True, values=None):
        super().__init__(attributes=attributes, *children)
        if fill:
            if name is not None:
                self.attributes["table:name"] = name
            self.children.append(GeneralElement("table:table-column"))
            if values is not None:
                for row in values:
                    self.append_row(row)

            if len(self.rows) == 0:
                raise OdioException("There must be at least one row")

    def append_row(self, row):
        row_element = Row()
        self.children.append(row_element)

        for val in row:
            row_element.append_cell(val)

    @property
    def rows(self):
        return self.find_descendants_by_name("table:table-row")

    @property
    def name(self):
        return self.attributes["table:name"]

    @property
    def values(self):
        return [row.values for row in self.rows]


class NumberText(BaseElement):
    tag_name = "number:text"
    allowed_types = (str,)

    def __init__(self, *children, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)


class NumberYear(BaseElement):
    tag_name = "number:year"

    def __init__(self, *children, is_long=True, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["number:style"] = "long" if is_long else "short"


class NumberMonth(BaseElement):
    tag_name = "number:month"

    def __init__(self, *children, is_long=True, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["number:style"] = "long" if is_long else "short"


class NumberDay(BaseElement):
    tag_name = "number:day"

    def __init__(self, *children, is_long=True, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["number:style"] = "long" if is_long else "short"


class NumberHours(BaseElement):
    tag_name = "number:hours"

    def __init__(self, *children, is_long=True, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["number:style"] = "long" if is_long else "short"


class NumberMinutes(BaseElement):
    tag_name = "number:minutes"

    def __init__(self, *children, is_long=True, attributes=None, fill=True):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["number:style"] = "long" if is_long else "short"


class DateStyle(BaseElement):
    tag_name = "number:date-style"

    def __init__(
        self,
        *children,
        name="date",
        year_long=True,
        year_month_sep="-",
        month_long=True,
        month_day_sep="-",
        day_long=True,
        day_hour_sep=" ",
        hour_long=True,
        hour_minute_sep=":",
        minute_long=True,
        attributes=None,
        fill=True,
    ):
        super().__init__(*children, attributes=attributes)
        if fill:
            self.attributes["style:name"] = name
            self.children.append(NumberYear(is_long=year_long))
            self.children.append(NumberText(year_month_sep))
            self.children.append(NumberMonth(is_long=month_long))
            self.children.append(NumberText(month_day_sep))
            self.children.append(NumberDay(is_long=day_long))
            self.children.append(NumberText(day_hour_sep))
            self.children.append(NumberHours(is_long=hour_long))
            self.children.append(NumberText(hour_minute_sep))
            self.children.append(NumberMinutes(is_long=minute_long))

    @property
    def name(self):
        return self.attributes["style:name"]


element_classes = {
    v.tag_name: v for k, v in globals().items() if hasattr(v, "tag_name")
}

for k, v in MAP.items():
    props = deepcopy(v)
    props["cls"] = element_classes.get(k)
    ELEMENT_MAP[k] = props



from io import BytesIO

from odio import parse_document
from odio.v1_3 import create_spreadsheet


def test_parse_document():
    sheet = create_spreadsheet()
    with BytesIO() as f:
        sheet.save(f)
        f.seek(0)
        parse_document(f)



from datetime import date as Date, datetime as Datetime
from io import BytesIO
from zipfile import ZipFile

from odio.v1_3 import create_spreadsheet, create_text, parse_document
from odio.v1_3.elements import (
    A,
    Cell,
    DateStyle,
    ELEMENT_MAP,
    Formula,
    H1,
    Manifest,
    ManifestFileEntry,
    P,
    Row,
    Span,
    Spreadsheet,
    Table,
)


def test_ELEMENT_MAP():
    for element_name, element_properties in ELEMENT_MAP.items():
        cls = element_properties["cls"]
        assert cls.tag_name == element_name


def test_create_text():
    text = create_text()
    doc = text.doc
    document_content = doc.document_content

    body_element = document_content.find_children_by_name("office:body")[0]
    body_element.find_children_by_name("office:text")[0]


def test_A_str():
    a = A("A link", href="https://example.com")

    expected = (
        '<text:a xlink:href="https://example.com" '
        'xlink:type="simple">A link</text:a>'
    )
    assert str(a) == expected


def test_A_href():
    href = href = "https://example.com"
    a = A("A link", href=href)

    assert a.href == href


def test_DateStyle():
    date_style = DateStyle()
    assert date_style.name == "date"


def test_Row_cells():
    v1 = 1
    v2 = 2
    v3 = Formula("=A1+A2")
    row = Row(Cell(value=v1), Cell(value=v2), Cell(value=v3))
    assert [cell.value for cell in row.cells] == [v1, v2, v3]


def test_Row_pack():
    row = Row(Cell(value=1), Cell(value=1))
    packed_row = row.pack()

    assert len(packed_row.children) == 1


def test_Cell():
    cell = Cell()
    assert cell.tag_name == "table:table-cell"


def test_Cell_date():
    dt = Datetime(2022, 6, 12)
    cell = Cell(value=dt)

    assert cell.value_type == "date"
    assert cell.attributes["office:date-value"] == "2022-06-12T00:00:00"
    assert cell.value == dt
    assert cell.tag_name == "table:table-cell"


def test_Cell_get_value_short_date():
    attributes = {
        "office:value-type": "date",
        "office:date-value": "2022-06-22",
    }
    cell = Cell(attributes=attributes, fill=False)

    assert cell.value == Date(2022, 6, 22)


def test_Cell_string():
    value = "atom"
    cell = Cell(value=value)
    assert cell.value == value


def test_Cell_formula():
    value = Formula("=B1 + C1")
    cell = Cell(value=value)
    assert cell.value == value


def test_H1():
    H1()


def test_ManifestFileEntry():
    full_path = "diary.txt"
    media_type = "text/plain"
    mfe = ManifestFileEntry(full_path=full_path, media_type=media_type)

    assert mfe.attributes["manifest:full-path"] == full_path
    assert mfe.full_path == full_path
    assert mfe.attributes["manifest:media-type"] == media_type
    assert mfe.media_type == media_type


def test_ManifestFileEntry_fill_False():
    ManifestFileEntry(fill=False)


def test_Manifest():
    media_type = "text/plain"
    m = Manifest(media_type=media_type)

    mfe = m.children[0]
    assert mfe.attributes["manifest:full-path"] == "/"
    assert mfe.attributes["manifest:media-type"] == media_type


def test_P():
    P()


def test_Row_append_cell():
    row = Row()
    row.append_cell(1)


def test_Span():
    Span()


def test_create_spreadsheet():
    spreadsheet = create_spreadsheet()
    doc = spreadsheet.doc
    document_content = doc.document_content

    body_element = document_content.find_children_by_name("office:body")[0]
    body_element.find_children_by_name("office:spreadsheet")[0]


def test_Spreadsheet_save():
    spreadsheet = create_spreadsheet()
    with BytesIO() as f:
        spreadsheet.save(f)


def test_Spreadsheet_append_table():
    spreadsheet = Spreadsheet()
    table_name = "a"
    table = spreadsheet.append_table(table_name, values=[[]])
    assert table.attributes["table:name"] == table_name


def test_Spreadsheet_append_table_with_rows():
    spreadsheet = Spreadsheet()
    table_name = "a"
    rows = [
        [2.1, "2"],
    ]
    table = spreadsheet.append_table(table_name, rows)
    assert table.attributes["table:name"] == table_name
    assert table.values == rows


def test_Table_init():
    table_name = "a"
    table = Table(name=table_name, values=[[]])
    assert table.attributes["table:name"] == table_name
    assert table.name == table_name
    assert table.values == [[]]


def test_Table_append_row():
    table = Table(name="Plan", values=[[]])
    table.append_row([1])


def test_parse_document():
    sheet = create_spreadsheet()
    with BytesIO() as f:
        sheet.save(f)

        f.seek(0)
        with ZipFile(f) as z:
            parse_document(z)



import datetime
import os
from zipfile import ZIP_STORED, ZipFile

from odio.v1_3 import Cell, Formula, create_spreadsheet, parse_document


def normalized_walk(path):
    return list(
        [os.path.relpath(pth, path), sorted(dirs), sorted(fls)]
        for pth, dirs, fls in sorted(os.walk(path))
    )


def test_create_parse_spreadsheet(tmpdir):
    TABLE_NAME = "Plan"
    py_1 = "veni, vidi, vici"
    py_2 = 0.3
    py_3 = py_4 = 5
    py_5 = Formula("=B1 + C1")
    py_6 = datetime.datetime(2015, 6, 30, 16, 38)
    py_7 = None
    py_8 = "Dombey & Son"
    py_9 = True
    ROW = [
        py_1,
        py_2,
        py_3,
        py_4,
        Cell(value=py_5),
        py_6,
        py_7,
        py_8,
        py_9,
    ]
    fname = tmpdir.join("actual.ods")
    sheet = create_spreadsheet()
    table = sheet.append_table(TABLE_NAME, values=[ROW])

    with open(str(fname), "wb") as f:
        sheet.save(f)
    actual_dir = str(tmpdir.mkdir("actual"))
    with ZipFile(str(fname)) as z:
        info = z.getinfo("mimetype")
        assert info.compress_type == ZIP_STORED
        z.extractall(actual_dir)

    desired_dir = str(os.path.join(os.path.dirname(__file__), "unpacked"))

    actual_walk = normalized_walk(actual_dir)
    desired_walk = normalized_walk(desired_dir)
    assert actual_walk == desired_walk

    for i, (actual_pth, actual_dirs, actual_fls) in enumerate(actual_walk):
        desired_pth, desired_dirs, desired_fls = desired_walk[i]
        for j, actual_fl in enumerate(actual_fls):
            desired_fl = desired_fls[j]
            with (
                open(os.path.join(actual_dir, actual_pth, actual_fl)) as actual_f,
                open(os.path.join(desired_dir, desired_pth, desired_fl)) as desired_f,
            ):
                ac = "".join(actual_f)
                de = "".join(desired_f)
                assert ac == de

    with open(str(fname), "rb") as f:
        with ZipFile(f) as z:
            sheet = parse_document(z)

    table = sheet.tables[0]
    assert table.name == TABLE_NAME
    expected_row = [
        py_1,
        py_2,
        py_3,
        py_4,
        py_5,
        py_6,
        py_7,
        py_8,
        py_9,
    ]
    actual_row = [c.value for c in table.rows[0].cells]
    assert actual_row == expected_row


def test_file_not_closed(tmpdir):
    fname = tmpdir.join("test.ods")
    f = open(str(fname), "wb")
    sheet = create_spreadsheet()
    sheet.append_table("Keats", values=[["Season", "of", "mists"]])
    sheet.save(f)
    f.seek(0)
    f.close()



from odio.v1_3.elements import (
    A,
    Cell,
    Document,
    ELEMENT_MAP,
    Formula,
    H,
    H1,
    H2,
    H3,
    H4,
    H5,
    H6,
    P,
    Span,
    Spreadsheet,
    Table,
    Text,
)
from odio.v1_3.xml import (
    GeneralElement,
    parse_xml_str,
)


def create_spreadsheet(date_style=None):
    doc = Document()
    spreadsheet = Spreadsheet(doc=doc)
    document_content = doc.document_content
    document_content.children.append(GeneralElement("office:scripts"))
    document_content.children.append(GeneralElement("office:automatic-styles"))
    document_styles = doc.document_styles
    styles = GeneralElement("office:styles")
    document_styles.children.append(styles)
    if date_style is not None:
        styles.children.append(date_style)
        styles.children.append(
            GeneralElement(
                "style:style",
                attributes={
                    "style:name": "cell_date",
                    "style:family": "table-cell",
                    "style:data-style-name": date_style.name,
                },
            ),
        )
    document_content.children.append(GeneralElement("office:body", spreadsheet))
    return spreadsheet


def create_text(date_style=None, *children):
    doc = Document()
    text = Text(doc=doc)
    document_content = doc.document_content
    document_content.children.append(GeneralElement("office:scripts"))
    document_content.children.append(GeneralElement("office:automatic-styles"))
    document_styles = doc.document_styles
    styles = GeneralElement("office:styles")
    document_styles.children.append(styles)
    if date_style is not None:
        styles.children.append(date_style)
        styles.children.append(
            GeneralElement(
                "style:style",
                attributes={
                    "style:name": "cell_date",
                    "style:family": "table-cell",
                    "style:data-style-name": date_style.name,
                },
            ),
        )
    document_content.children.append(GeneralElement("office:body", text))
    return text


def parse_document(z):
    manifest = parse_xml_str(ELEMENT_MAP, z.read("META-INF/manifest.xml"))
    kwargs = {}
    for entry in manifest.children:
        if entry.full_path == "/":
            continue
        element = parse_xml_str(ELEMENT_MAP, z.read(entry.full_path))
        kwarg = element.tag_name.split("-")[-1]
        kwargs[kwarg] = element
    doc = Document(**kwargs)
    return doc.document_content.body_root


__all__ = [
    "A",
    "Cell",
    "DateStyle",
    "Formula",
    "H",
    "H1",
    "H2",
    "H3",
    "H4",
    "H5",
    "H6",
    "P",
    "Span",
    "Table",
    "create_spreadsheet",
    "parse_spreadsheet",
]



import xml.dom.minidom
import zipfile


class OdioException(Exception):
    pass


def parse_document(f):
    with zipfile.ZipFile(f, "r") as z:
        dom = xml.dom.minidom.parseString(z.read("META-INF/manifest.xml"))
        version = dom.documentElement.getAttribute("manifest:version")

        if version == "1.1":
            import odio.v1_1

            return odio.v1_1.parser.parse_node(dom)
        elif version == "1.2":
            import odio.v1_2

            return odio.v1_2.parse_document(z)
        elif version == "1.3":
            import odio.v1_3

            return odio.v1_3.parse_document(z)
        else:
            raise Exception(
                f"The version '{version}' isn't recognized. The valid version strings "
                f"are '1.1', '1.2' and '1.3'."
            )



from io import StringIO
from xml.dom.minidom import parseString

import pytest

from odio.v1_2.xml import (
    Attributes,
    Children,
    Element,
    GeneralElement,
    OdioException,
    document_to_str,
    element_to_str,
    parse_element,
    parse_xml_str,
    write_element,
)


def test_Attributes_init():
    Attributes({})


def test_Attributes_init_v_none():
    with pytest.raises(
        OdioException,
        match="The type of an attribute value must be str. Found <class 'NoneType'>",
    ):
        Attributes({"version": None})


def test_Attributes_setitem_v_none():
    attrs = Attributes()
    with pytest.raises(
        OdioException,
        match="The type of an attribute value must be str. Found <class 'NoneType'>",
    ):
        attrs["version"] = None


def test_Attributes_repr():
    repr(Attributes({"version": "1.2"}, allowed={"version"}))


def test_Children_check_type():
    with pytest.raises(
        OdioException,
        match="A child must be a str or a subclass of Element, but found a "
        "<class 'dict'>",
    ):
        Children([{}], allowed=(str,))


def test_Children_iadd():
    with pytest.raises(
        OdioException,
        match="A child must be a str or a subclass of Element, but found a "
        "<class 'int'>",
    ):
        children = Children([], allowed=(str,))
        children += [1]


def test_Children_append_mixed():
    children = Children([], allowed=("*",))
    children.append(" modesty and a ")
    children.append(" manly character ")
    assert children == [" modesty and a ", " manly character "]


def test_Children_append_cdata():
    pass


def test_Children_append_elements():
    pass


def test_GeneralElement_find_descendants_by_name():
    element = GeneralElement("table")
    assert element.find_descendants_by_name("row") == []


def test_write_element_simple():
    element = GeneralElement("spreadsheet")
    with StringIO() as f:
        write_element(f, element)
        assert f.getvalue() == "<spreadsheet/>"


def test_write_element_attributes():
    element = GeneralElement("spreadsheet", attributes={"version": "1.2"})
    with StringIO() as f:
        write_element(f, element)
        assert f.getvalue() == '<spreadsheet version="1.2"/>'


def test_write_element():
    element = GeneralElement("spreadsheet", GeneralElement("table"))
    with StringIO() as f:
        write_element(f, element)
        expected = "<spreadsheet><table/></spreadsheet>"
        assert f.getvalue() == expected


def test_write_element_single_str_child():
    element = GeneralElement("spreadsheet", "*")
    with StringIO() as f:
        write_element(f, element)
        expected = "<spreadsheet>*</spreadsheet>"
        assert f.getvalue() == expected


def test_GeneralElement_eq():
    a = GeneralElement("a")
    b = GeneralElement("a")
    assert a == b


def test_GeneralElement_find_children_by_name():
    spreadsheet = GeneralElement("spreadsheet")
    element = GeneralElement("document", spreadsheet, GeneralElement("text"))

    children = element.find_children_by_name("spreadsheet")

    assert children == [spreadsheet]


def test_GeneralElement_to_text():
    element = GeneralElement("cell", "")
    assert element.to_text() == ""


def test_GeneralElement_child_type():
    with pytest.raises(
        OdioException,
        match="A child must be a str or a subclass of Element, but found a "
        "<class 'dict'>",
    ):
        GeneralElement("table", {"name": "Alf"})


def test_document_to_str():
    element = GeneralElement("spreadsheet")
    document_to_str(element)


def test_element_to_str():
    element = GeneralElement("spreadsheet")
    element_to_str(element)


def test_element_to_str_text():
    element = GeneralElement(
        "p",
        "From my grandfather Verus: the lessons of noble character ",
        "and even temper.",
    )
    actual = element_to_str(element)
    assert actual == (
        "<p>From my grandfather Verus: the lessons of noble character and even "
        "temper.</p>"
    )


def test_element_to_str_mixed():
    element = GeneralElement(
        "p",
        "From my grandfather Verus: the lessons of noble character ",
        GeneralElement("em", "and"),
        " even temper.",
    )
    actual = element_to_str(element)
    assert actual == (
        "<p>From my grandfather Verus: the lessons of noble character <em>and</em> "
        "even temper.</p>"
    )


def test_parse_element_one_str_child():
    xml_str = """<?xml version="1.0" encoding="UTF-8"?>
<cell> </cell>"""
    dom = parseString(xml_str)

    class Cell(Element):
        tag_name = "cell"

        def __init__(self, *children, attributes=None, fill=True):
            super().__init__(
                *children,
                attributes=attributes,
                allowed_attributes={},
                allowed_children={str},
            )

    element_map = {
        "cell": {
            "allowed_children": {str},
            "allowed_attributes": {},
            "cls": Cell,
        }
    }
    element = parse_element(element_map, dom.documentElement)
    assert element.children[0] == " "


def test_parse_xml_str():
    element_map = {}
    xml_str = """<?xml version="1.0" encoding="UTF-8"?>
  <p>electron</p>"""
    p = parse_xml_str(element_map, xml_str)
    assert p.children[0] == "electron"


def test_parse_xml_str_unpack():
    class Ul(Element):
        tag_name = "ul"

        def __init__(self, *children, attributes=None, fill=True):
            super().__init__(
                *children,
                attributes=attributes,
                allowed_attributes={},
                allowed_children={"li"},
            )

        def unpack(self):
            self.children.append(GeneralElement("li"))
            return self

    element_map = {"ul": {"cls": Ul}}
    xml_str = """<?xml version="1.0" encoding="UTF-8"?>
  <ul>
    <li>electron</li>
  </ul>"""
    ul = parse_xml_str(element_map, xml_str)
    assert len(ul.children) == 2



import pytest

from odio import OdioException
from odio.v1_1 import create_spreadsheet


def test_writerow(tmpdir):
    with pytest.raises(
        OdioException,
        match="Version '1.1' isn't supported for creating spreadsheets. Use versions "
        "1.3 or 1.2 instead.",
    ):
        create_spreadsheet()



from types import MappingProxyType

EM = {
    "office:document-content": {
        "allowed_attributes": {
            "grddl:transformation",
            "office:version",
            "xmlns:anim",
            "xmlns:chart",
            "xmlns:config",
            "xmlns:db",
            "xmlns:dr3d",
            "xmlns:draw",
            "xmlns:form",
            "xmlns:meta",
            "xmlns:number",
            "xmlns:office",
            "xmlns:presentation",
            "xmlns:script",
            "xmlns:table",
            "xmlns:text",
            "xmlns:style",
            "xmlns:fo",
            "xmlns:svg",
            "xmlns:smil",
            "xmlns:dc",
            "xmlns:math",
            "xmlns:xforms",
            "xmlns:xlink",
            "xmlns:xhtml",
            "xmlns:grddl",
            "xmlns:of",
        },
        "allowed_children": {
            "office:automatic-styles",
            "office:body",
            "office:font-face-decls",
            "office:scripts",
        },
    },
    "office:document-meta": {
        "allowed_attributes": {
            "grddl:transformation",
            "office:version",
            "xmlns:office",
            "xmlns:xlink",
            "xmlns:dc",
            "xmlns:meta",
            "xmlns:of",
        },
        "allowed_children": {"office:meta"},
    },
    "office:document-settings": {
        "allowed_attributes": {
            "grddl:transformation",
            "office:version",
            "xmlns:office",
            "xmlns:xlink",
            "xmlns:config",
            "xmlns:of",
        },
        "allowed_children": {"office:settings"},
    },
    "office:document-styles": {
        "allowed_attributes": {
            "grddl:transformation",
            "office:version",
            "xmlns:anim",
            "xmlns:chart",
            "xmlns:config",
            "xmlns:db",
            "xmlns:dr3d",
            "xmlns:draw",
            "xmlns:form",
            "xmlns:meta",
            "xmlns:number",
            "xmlns:office",
            "xmlns:presentation",
            "xmlns:script",
            "xmlns:table",
            "xmlns:text",
            "xmlns:style",
            "xmlns:odf",
            "xmlns:fo",
            "xmlns:svg",
            "xmlns:smil",
            "xmlns:dc",
            "xmlns:math",
            "xmlns:xforms",
            "xmlns:xlink",
            "xmlns:xhtml",
            "xmlns:grddl",
            "xmlns:pkg",
            "xmlns:of",
        },
        "allowed_children": {
            "office:automatic-styles",
            "office:font-face-decls",
            "office:master-styles",
            "office:styles",
        },
    },
    "office:spreadsheet": {
        "allowed_attributes": {
            "table:protection-key",
            "table:protection-key-digest-algorithm",
            "table:structure-protected",
        },
        "allowed_children": {
            "table:calculation-settings",
            "table:consolidation",
            "table:content-validations",
            "table:database-ranges",
            "table:data-pilot-tables",
            "table:dde-links",
            "table:label-ranges",
            "table:named-expressions",
            "table:table",
            "table:tracked-changes",
            "text:alphabetical-index-auto-mark-file",
            "text:dde-connection-decls",
            "text:sequence-decls",
            "text:user-field-decls",
            "text:variable-decls",
        },
    },
    "office:text": {
        "allowed_attributes": {"text:global", "text:use-soft-page-breaks"},
        "allowed_children": {
            "dr3d:scene",
            "draw:a",
            "draw:caption",
            "draw:circle",
            "draw:connector",
            "draw:control",
            "draw:custom-shape",
            "draw:ellipse",
            "draw:frame",
            "draw:g",
            "draw:line",
            "draw:measure",
            "draw:page-thumbnail",
            "draw:path",
            "draw:polygon",
            "draw:polyline",
            "draw:rect",
            "draw:regular-polygon",
            "office:forms",
            "table:calculation-settings",
            "table:consolidation",
            "table:content-validations",
            "table:database-ranges",
            "table:data-pilot-tables",
            "table:dde-links",
            "table:label-ranges",
            "table:named-expressions",
            "table:table",
            "text:alphabetical-index",
            "text:alphabetical-index-auto-mark-file",
            "text:bibliography",
            "text:change",
            "text:change-end",
            "text:change-start",
            "text:dde-connection-decls",
            "text:h",
            "text:illustration-index",
            "text:list",
            "text:numbered-paragraph",
            "text:object-index",
            "text:p",
            "text:page-sequence",
            "text:section",
            "text:sequence-decls",
            "text:soft-page-break",
            "text:table-index",
            "text:table-of-content",
            "text:tracked-changes",
            "text:user-field-decls",
            "text:user-index",
            "text:variable-decls",
        },
    },
    "manifest:file-entry": {
        "allowed_attributes": {
            "manifest:full-path",
            "manifest:media-type",
            "manifest:preferred-view-mode",
            "manifest:size",
            "manifest:version",
        },
        "allowed_children": {"manifest:encryption-data"},
    },
    "manifest:manifest": {
        "allowed_attributes": {"manifest:version", "xmlns:manifest"},
        "allowed_children": {"manifest:file-entry"},
    },
    "number:date-style": {
        "allowed_attributes": {
            "number:automatic-order",
            "number:country",
            "number:format-source",
            "number:language",
            "number:rfc-language-tag",
            "number:script",
            "number:title",
            "number:transliteration-country",
            "number:transliteration-format",
            "number:transliteration-language",
            "number:transliteration-style",
            "style:display-name",
            "style:name",
            "style:volatile",
        },
        "allowed_children": {
            "number:am-pm",
            "number:day",
            "number:day-of-week",
            "number:era",
            "number:hours",
            "number:minutes",
            "number:month",
            "number:quarter",
            "number:seconds",
            "number:text",
            "number:week-of-year",
            "number:year",
            "style:map",
            "style:text-properties",
        },
    },
    "number:year": {
        "allowed_attributes": {"number:calendar", "number:style"},
        "allowed_children": {},
    },
    "number:text": {
        "allowed_attributes": {},
        "allowed_children": {str},
    },
    "number:month": {
        "allowed_attributes": {
            "number:calendar",
            "number:possessive-form",
            "number:style",
            "number:textual",
        },
        "allowed_children": {},
    },
    "number:day": {
        "allowed_attributes": {
            "number:calendar",
            "number:style",
        },
        "allowed_children": {},
    },
    "number:hours": {
        "allowed_attributes": {
            "number:style",
        },
        "allowed_children": {},
    },
    "number:minutes": {
        "allowed_attributes": {
            "number:style",
        },
        "allowed_children": {},
    },
    "table:table": {
        "allowed_attributes": {
            "table:is-sub-table",
            "table:name",
            "table:print",
            "table:print-ranges",
            "table:protected",
            "table:protection-key",
            "table:protection-key-digest-algorithm",
            "table:style-name",
            "table:template-name",
            "table:use-banding-columns-styles",
            "table:use-banding-rows-styles",
            "table:use-first-column-styles",
            "table:use-first-row-styles",
            "table:use-last-column-styles",
            "table:use-last-row-styles",
            "xml:id",
        },
        "allowed_children": {
            "office:dde-source",
            "office:forms",
            "table:desc",
            "table:named-expressions",
            "table:scenario",
            "table:shapes",
            "table:table-column",
            "table:table-column-group",
            "table:table-columns",
            "table:table-header-columns",
            "table:table-header-rows",
            "table:table-row",
            "table:table-row-group",
            "table:table-rows",
            "table:table-source",
            "table:title",
            "text:soft-page-break",
        },
    },
    "table:table-row": {
        "allowed_attributes": {
            "table:default-cell-style-name",
            "table:number-rows-repeated",
            "table:style-name",
            "table:visibility",
            "xml:id",
        },
        "allowed_children": {"table:covered-table-cell", "table:table-cell"},
    },
    "text:a": {
        "allowed_attributes": {
            "office:name",
            "office:target-frame-name",
            "office:title",
            "text:style-name",
            "text:visited-style-name",
            "xlink:actuate",
            "xlink:href",
            "xlink:show",
            "xlink:type",
        },
        "allowed_children": {
            str,
            "dr3d:scene",
            "draw:a",
            "draw:caption",
            "draw:circle",
            "draw:connector",
            "draw:control",
            "draw:custom-shape",
            "draw:ellipse",
            "draw:frame",
            "draw:g",
            "draw:line",
            "draw:measure",
            "draw:page-thumbnail",
            "draw:path",
            "draw:polygon",
            "draw:polyline",
            "draw:rect",
            "draw:regular-polygon",
            "office:annotation",
            "office:annotation-end",
            "office:event-listeners",
            "presentation:date-time",
            "presentation:footer",
            "presentation:header",
            "text:alphabetical-index-mark",
            "text:alphabetical-index-mark-end",
            "text:alphabetical-index-mark-start",
            "text:author-initials",
            "text:author-name",
            "text:bibliography-mark",
            "text:bookmark",
            "text:bookmark-end",
            "text:bookmark-ref",
            "text:bookmark-start",
            "text:change",
            "text:change-end",
            "text:change-start",
            "text:chapter",
            "text:character-count",
            "text:conditional-text",
            "text:creation-date",
            "text:creation-time",
            "text:creator",
            "text:database-display",
            "text:database-name",
            "text:database-next",
            "text:database-row-number",
            "text:database-row-select",
            "text:date",
            "text:dde-connection",
            "text:description",
            "text:editing-cycles",
            "text:editing-duration",
            "text:execute-macro",
            "text:expression",
            "text:file-name",
            "text:hidden-paragraph",
            "text:hidden-text",
            "text:image-count",
            "text:initial-creator",
            "text:keywords",
            "text:line-break",
            "text:measure",
            "text:meta",
            "text:meta-field",
            "text:modification-date",
            "text:modification-time",
            "text:note",
            "text:note-ref",
            "text:object-count",
            "text:page-continuation",
            "text:page-count",
            "text:page-number",
            "text:page-variable-get",
            "text:page-variable-set",
            "text:paragraph-count",
            "text:placeholder",
            "text:print-date",
            "text:printed-by",
            "text:print-time",
            "text:reference-mark",
            "text:reference-mark-end",
            "text:reference-mark-start",
            "text:reference-ref",
            "text:ruby",
            "text:s",
            "text:script",
            "text:sender-city",
            "text:sender-company",
            "text:sender-country",
            "text:sender-email",
            "text:sender-fax",
            "text:sender-firstname",
            "text:sender-initials",
            "text:sender-lastname",
            "text:sender-phone-private",
            "text:sender-phone-work",
            "text:sender-position",
            "text:sender-postal-code",
            "text:sender-state-or-province",
            "text:sender-street",
            "text:sender-title",
            "text:sequence",
            "text:sequence-ref",
            "text:sheet-name",
            "text:soft-page-break",
            "text:span",
            "text:subject",
            "text:tab",
            "text:table-count",
            "text:table-formula",
            "text:template-name",
            "text:text-input",
            "text:time",
            "text:title",
            "text:toc-mark",
            "text:toc-mark-end",
            "text:toc-mark-start",
            "text:user-defined",
            "text:user-field-get",
            "text:user-field-input",
            "text:user-index-mark",
            "text:user-index-mark-end",
            "text:user-index-mark-start",
            "text:variable-get",
            "text:variable-input",
            "text:variable-set",
            "text:word-count",
        },
    },
    "table:table-cell": {
        "allowed_attributes": {
            "office:boolean-value",
            "office:currency",
            "office:date-value",
            "office:string-value",
            "office:time-value",
            "office:value",
            "office:value-type",
            "table:content-validation-name",
            "table:formula",
            "table:number-columns-repeated",
            "table:number-columns-spanned",
            "table:number-matrix-columns-spanned",
            "table:number-matrix-rows-spanned",
            "table:number-rows-spanned",
            "table:protect",
            "table:protected",
            "table:style-name",
            "xhtml:about",
            "xhtml:content",
            "xhtml:datatype",
            "xhtml:property",
            "xml:id",
        },
        "allowed_children": {
            "dr3d:scene",
            "draw:a",
            "draw:caption",
            "draw:circle",
            "draw:connector",
            "draw:control",
            "draw:custom-shape",
            "draw:ellipse",
            "draw:frame",
            "draw:g",
            "draw:line",
            "draw:measure",
            "draw:page-thumbnail",
            "draw:path",
            "draw:polygon",
            "draw:polyline",
            "draw:rect",
            "draw:regular-polygon",
            "office:annotation",
            "table:cell-range-source",
            "table:detective",
            "table:table",
            "text:alphabetical-index",
            "text:bibliography",
            "text:change",
            "text:change-end",
            "text:change-start",
            "text:h",
            "text:illustration-index",
            "text:list",
            "text:numbered-paragraph",
            "text:object-index",
            "text:p",
            "text:section",
            "text:soft-page-break",
            "text:table-index",
            "text:table-of-content",
            "text:user-index",
        },
    },
    "text:h": {
        "allowed_attributes": {
            "text:class-names",
            "text:cond-style-name",
            "text:id",
            "text:is-list-header",
            "text:outline-level",
            "text:restart-numbering",
            "text:start-value",
            "text:style-name",
            "xhtml:about",
            "xhtml:content",
            "xhtml:datatype",
            "xhtml:property",
            "xml:id",
        },
        "allowed_children": {
            str,
            "dr3d:scene",
            "draw:a",
            "draw:caption",
            "draw:circle",
            "draw:connector",
            "draw:control",
            "draw:custom-shape",
            "draw:ellipse",
            "draw:frame",
            "draw:g",
            "draw:line",
            "draw:measure",
            "draw:page-thumbnail",
            "draw:path",
            "draw:polygon",
            "draw:polyline",
            "draw:rect",
            "draw:regular-polygon",
            "office:annotation",
            "office:annotation-end",
            "presentation:date-time",
            "presentation:footer",
            "presentation:header",
            "text:a",
            "text:alphabetical-index-mark",
            "text:alphabetical-index-mark-end",
            "text:alphabetical-index-mark-start",
            "text:author-initials",
            "text:author-name",
            "text:bibliography-mark",
            "text:bookmark",
            "text:bookmark-end",
            "text:bookmark-ref",
            "text:bookmark-start",
            "text:change",
            "text:change-end",
            "text:change-start",
            "text:chapter",
            "text:character-count",
            "text:conditional-text",
            "text:creation-date",
            "text:creation-time",
            "text:creator",
            "text:database-display",
            "text:database-name",
            "text:database-next",
            "text:database-row-number",
            "text:database-row-select",
            "text:date",
            "text:dde-connection",
            "text:description",
            "text:editing-cycles",
            "text:editing-duration",
            "text:execute-macro",
            "text:expression",
            "text:file-name",
            "text:hidden-paragraph",
            "text:hidden-text",
            "text:image-count",
            "text:initial-creator",
            "text:keywords",
            "text:line-break",
            "text:measure",
            "text:meta",
            "text:meta-field",
            "text:modification-date",
            "text:modification-time",
            "text:note",
            "text:note-ref",
            "text:number",
            "text:object-count",
            "text:page-continuation",
            "text:page-count",
            "text:page-number",
            "text:page-variable-get",
            "text:page-variable-set",
            "text:paragraph-count",
            "text:placeholder",
            "text:print-date",
            "text:printed-by",
            "text:print-time",
            "text:reference-mark",
            "text:reference-mark-end",
            "text:reference-mark-start",
            "text:reference-ref",
            "text:ruby",
            "text:s",
            "text:script",
            "text:sender-city",
            "text:sender-company",
            "text:sender-country",
            "text:sender-email",
            "text:sender-fax",
            "text:sender-firstname",
            "text:sender-initials",
            "text:sender-lastname",
            "text:sender-phone-private",
            "text:sender-phone-work",
            "text:sender-position",
            "text:sender-postal-code",
            "text:sender-state-or-province",
            "text:sender-street",
            "text:sender-title",
            "text:sequence",
            "text:sequence-ref",
            "text:sheet-name",
            "text:soft-page-break",
            "text:span",
            "text:subject",
            "text:tab",
            "text:table-count",
            "text:table-formula",
            "text:template-name",
            "text:text-input",
            "text:time",
            "text:title",
            "text:toc-mark",
            "text:toc-mark-end",
            "text:toc-mark-start",
            "text:user-defined",
            "text:user-field-get",
            "text:user-field-input",
            "text:user-index-mark",
            "text:user-index-mark-end",
            "text:user-index-mark-start",
            "text:variable-get",
            "text:variable-input",
            "text:variable-set",
            "text:word-count",
        },
    },
    "text:p": {
        "allowed_attributes": {
            "text:class-names",
            "text:cond-style-name",
            "text:id",
            "text:style-name",
            "xhtml:about",
            "xhtml:content",
            "xhtml:datatype",
            "xhtml:property",
            "xml:id",
        },
        "allowed_children": {
            str,
            "dr3d:scene",
            "draw:a",
            "draw:caption",
            "draw:circle",
            "draw:connector",
            "draw:control",
            "draw:custom-shape",
            "draw:ellipse",
            "draw:frame",
            "draw:g",
            "draw:line",
            "draw:measure",
            "draw:page-thumbnail",
            "draw:path",
            "draw:polygon",
            "draw:polyline",
            "draw:rect",
            "draw:regular-polygon",
            "office:annotation",
            "office:annotation-end",
            "presentation:date-time",
            "presentation:footer",
            "presentation:header",
            "text:a",
            "text:alphabetical-index-mark",
            "text:alphabetical-index-mark-end",
            "text:alphabetical-index-mark-start",
            "text:author-initials",
            "text:author-name",
            "text:bibliography-mark",
            "text:bookmark",
            "text:bookmark-end",
            "text:bookmark-ref",
            "text:bookmark-start",
            "text:change",
            "text:change-end",
            "text:change-start",
            "text:chapter",
            "text:character-count",
            "text:conditional-text",
            "text:creation-date",
            "text:creation-time",
            "text:creator",
            "text:database-display",
            "text:database-name",
            "text:database-next",
            "text:database-row-number",
            "text:database-row-select",
            "text:date",
            "text:dde-connection",
            "text:description",
            "text:editing-cycles",
            "text:editing-duration",
            "text:execute-macro",
            "text:expression",
            "text:file-name",
            "text:hidden-paragraph",
            "text:hidden-text",
            "text:image-count",
            "text:initial-creator",
            "text:keywords",
            "text:line-break",
            "text:measure",
            "text:meta",
            "text:meta-field",
            "text:modification-date",
            "text:modification-time",
            "text:note",
            "text:note-ref",
            "text:object-count",
            "text:page-continuation",
            "text:page-count",
            "text:page-number",
            "text:page-variable-get",
            "text:page-variable-set",
            "text:paragraph-count",
            "text:placeholder",
            "text:print-date",
            "text:printed-by",
            "text:print-time",
            "text:reference-mark",
            "text:reference-mark-end",
            "text:reference-mark-start",
            "text:reference-ref",
            "text:ruby",
            "text:s",
            "text:script",
            "text:sender-city",
            "text:sender-company",
            "text:sender-country",
            "text:sender-email",
            "text:sender-fax",
            "text:sender-firstname",
            "text:sender-initials",
            "text:sender-lastname",
            "text:sender-phone-private",
            "text:sender-phone-work",
            "text:sender-position",
            "text:sender-postal-code",
            "text:sender-state-or-province",
            "text:sender-street",
            "text:sender-title",
            "text:sequence",
            "text:sequence-ref",
            "text:sheet-name",
            "text:soft-page-break",
            "text:span",
            "text:subject",
            "text:tab",
            "text:table-count",
            "text:table-formula",
            "text:template-name",
            "text:text-input",
            "text:time",
            "text:title",
            "text:toc-mark",
            "text:toc-mark-end",
            "text:toc-mark-start",
            "text:user-defined",
            "text:user-field-get",
            "text:user-field-input",
            "text:user-index-mark",
            "text:user-index-mark-end",
            "text:user-index-mark-start",
            "text:variable-get",
            "text:variable-input",
            "text:variable-set",
            "text:word-count",
        },
    },
    "text:span": {
        "allowed_attributes": {"text:class-names", "text:style-name"},
        "allowed_children": {
            str,
            "dr3d:scene",
            "draw:a",
            "draw:caption",
            "draw:circle",
            "draw:connector",
            "draw:control",
            "draw:custom-shape",
            "draw:ellipse",
            "draw:frame",
            "draw:g",
            "draw:line",
            "draw:measure",
            "draw:page-thumbnail",
            "draw:path",
            "draw:polygon",
            "draw:polyline",
            "draw:rect",
            "draw:regular-polygon",
            "office:annotation",
            "office:annotation-end",
            "presentation:date-time",
            "presentation:footer",
            "presentation:header",
            "text:a",
            "text:alphabetical-index-mark",
            "text:alphabetical-index-mark-end",
            "text:alphabetical-index-mark-start",
            "text:author-initials",
            "text:author-name",
            "text:bibliography-mark",
            "text:bookmark",
            "text:bookmark-end",
            "text:bookmark-ref",
            "text:bookmark-start",
            "text:change",
            "text:change-end",
            "text:change-start",
            "text:chapter",
            "text:character-count",
            "text:conditional-text",
            "text:creation-date",
            "text:creation-time",
            "text:creator",
            "text:database-display",
            "text:database-name",
            "text:database-next",
            "text:database-row-number",
            "text:database-row-select",
            "text:date",
            "text:dde-connection",
            "text:description",
            "text:editing-cycles",
            "text:editing-duration",
            "text:execute-macro",
            "text:expression",
            "text:file-name",
            "text:hidden-paragraph",
            "text:hidden-text",
            "text:image-count",
            "text:initial-creator",
            "text:keywords",
            "text:line-break",
            "text:measure",
            "text:meta",
            "text:meta-field",
            "text:modification-date",
            "text:modification-time",
            "text:note",
            "text:note-ref",
            "text:object-count",
            "text:page-continuation",
            "text:page-count",
            "text:page-number",
            "text:page-variable-get",
            "text:page-variable-set",
            "text:paragraph-count",
            "text:placeholder",
            "text:print-date",
            "text:printed-by",
            "text:print-time",
            "text:reference-mark",
            "text:reference-mark-end",
            "text:reference-mark-start",
            "text:reference-ref",
            "text:ruby",
            "text:s",
            "text:script",
            "text:sender-city",
            "text:sender-company",
            "text:sender-country",
            "text:sender-email",
            "text:sender-fax",
            "text:sender-firstname",
            "text:sender-initials",
            "text:sender-lastname",
            "text:sender-phone-private",
            "text:sender-phone-work",
            "text:sender-position",
            "text:sender-postal-code",
            "text:sender-state-or-province",
            "text:sender-street",
            "text:sender-title",
            "text:sequence",
            "text:sequence-ref",
            "text:sheet-name",
            "text:soft-page-break",
            "text:span",
            "text:subject",
            "text:tab",
            "text:table-count",
            "text:table-formula",
            "text:template-name",
            "text:text-input",
            "text:time",
            "text:title",
            "text:toc-mark",
            "text:toc-mark-end",
            "text:toc-mark-start",
            "text:user-defined",
            "text:user-field-get",
            "text:user-field-input",
            "text:user-index-mark",
            "text:user-index-mark-end",
            "text:user-index-mark-start",
            "text:variable-get",
            "text:variable-input",
            "text:variable-set",
            "text:word-count",
        },
    },
}

MAP = MappingProxyType(EM)



from abc import ABC
from collections import UserDict, UserList
from io import StringIO
from xml.dom import Node
from xml.dom.minidom import parseString


class OdioException(Exception):
    pass


class Children(UserList):
    def __init__(self, iterable=None, *, allowed=()):
        self.allowed = allowed
        super().__init__()
        if iterable is not None:
            for v in iterable:
                self.append(v)

    """
    "__iadd__", "__imul__", "__init__", "__init_subclass__", "__iter__", "__le__",
    "__len__", "__lt__", "__mul__", "__ne__", "__new__", "__reduce__", "__reduce_ex__",
    "__repr__", "__reversed__", "__rmul__", "__setattr__", "__setitem__", "__sizeof__",
    "__str__", "__subclasshook__", "append", "clear", "copy", "count", "extend",
    "index", "insert", "pop", "remove", "reverse", "sort"
    """

    def __iadd__(self, other):
        for v in other:
            self.append(v)

    def __delitem__(self):
        raise OdioException("Children are append-only in Odio.")

    def __setitem__(self, val):
        raise OdioException("Children are append-only in Odio.")

    def append(self, value):
        if len(self.allowed) == 0:
            raise OdioException("No children are allowed for this element.")

        if isinstance(value, str):
            if str not in self.allowed and "*" not in self.allowed:
                raise OdioException(
                    f"Children of type str aren't allowed for this element. The "
                    f"allowed children are {self.allowed}."
                )
            super().append(value)

        elif isinstance(value, Element):
            if value.tag_name not in self.allowed and "*" not in self.allowed:
                raise OdioException(
                    f"Children with tag name {value.tag_name} aren't allowed for this "
                    f"element. The allowed types of children are {self.allowed}."
                )
            else:
                super().append(value)
        else:
            raise OdioException(
                f"A child must be a str or a subclass of Element, but found a "
                f"{type(value)}"
            )


class Attributes(UserDict):
    def __init__(self, dictionary=None, allowed=None):
        self._allowed = {} if allowed is None else allowed
        super().__init__()
        if dictionary is not None:
            for k, v in dictionary.items():
                self[k] = v

    def __setitem__(self, k, v):
        self._check_pair(k, v)
        super().__setitem__(k, v)

    def _check_pair(self, k, v):
        if not isinstance(k, str):
            raise OdioException(
                f"The type of an attribute name must be str. Found {type(k)}"
            )
        if not isinstance(v, str):
            raise OdioException(
                f"The type of an attribute value must be str. Found {type(v)}"
            )

        if k not in self._allowed and "*" not in self._allowed:
            raise OdioException(
                f"The name of the attribute was {k} but it must be one of "
                f"{self._allowed}"
            )


class Element(ABC):
    def __init__(
        self, *children, attributes=None, allowed_attributes=None, allowed_children=None
    ):
        self.attributes = Attributes(attributes, allowed=allowed_attributes)
        self.children = Children(children, allowed=allowed_children)

    def __repr__(self):
        return f"odio.Element({self.tag_name})"

    def __eq__(self, other):
        return (
            isinstance(other, Element)
            and self.tag_name == other.tag_name
            and self.children == other.children
            and self.attributes == other.attributes
        )

    def to_text(self):
        return "".join(n for n in self.children if isinstance(n, str))

    def find_children_by_name(self, tag_name):
        return [c for c in self.children if c.tag_name == tag_name]

    def pack(self):
        return self

    def unpack(self):
        return self

    def find_descendants_by_name(self, tag_name):
        descendants = []
        for child in self.children:
            if isinstance(child, str):
                continue
            else:
                if child.tag_name == tag_name:
                    descendants.append(child)
                descendants.extend(child.find_descendants_by_name(tag_name))

        return descendants

    def __str__(self):
        return element_to_str(self)


class GeneralElement(Element):
    def __init__(self, tag_name, *children, attributes=None, fill=True):
        self.tag_name = tag_name
        allowed_attributes = {"*"}
        allowed_children = {"*"}
        super().__init__(
            *children,
            attributes=attributes,
            allowed_attributes=allowed_attributes,
            allowed_children=allowed_children,
        )


def escape(data):
    return data.replace("&", "&amp;").replace(">", "&gt;").replace("<", "&lt;")


def quoteattr(data):
    d = escape(data)

    if '"' in d:
        if "'" in d:
            return f'''"{d.replace('"', '&quot;')}"'''
        else:
            return f"'{d}'"
    else:
        return f'"{d}"'


def atts_to_str(attrs):
    if len(attrs) == 0:
        return ""
    else:
        return " " + " ".join(f"{k}={quoteattr(v)}" for k, v in sorted(attrs.items()))


def write_element(f, element):
    element = element.pack()
    tag_name = element.tag_name
    att_str = atts_to_str(element.attributes)
    children = element.children

    if len(children) == 0:
        f.write(f"<{tag_name}{att_str}/>")
    elif len(children) == 1 and isinstance(children[0], str):
        f.write(f"<{tag_name}{att_str}>{children[0]}</{tag_name}>")
    else:
        f.write(f"<{tag_name}{att_str}>")
        for child in children:
            if isinstance(child, str):
                f.write(escape(child))
            else:
                write_element(f, child)

        f.write(f"</{tag_name}>")


def document_to_str(document):
    with StringIO() as f:
        write_document(f, document)
        return f.getvalue()


def element_to_str(element):
    with StringIO() as f:
        write_element(f, element)
        return f.getvalue()


def write_document(f, element):
    f.write('<?xml version="1.0" encoding="utf-8"?>\n')
    write_element(f, element)
    f.write("\n")


def parse_xml_str(element_map, xml_str):
    dom = parseString(xml_str)
    return parse_element(element_map, dom.documentElement)


def parse_element(element_map, node):
    tag = node.tagName

    try:
        cls = element_map[tag]["cls"]
        try:
            element = cls(fill=False)
        except TypeError as e:
            raise OdioException(f"Problem with {cls}: {e}")
    except KeyError:
        try:
            element = GeneralElement(tag, fill=False)
        except TypeError as e:
            raise OdioException(f"Problem with {cls}: {e}")

    for i in range(node.attributes.length):
        attr = node.attributes.item(i)
        element.attributes[attr.name] = attr.value

    child_nodes = node.childNodes

    if str in element.children.allowed and len(element.children.allowed) == 1:
        element.children.append(child_nodes.item(0).nodeValue)
    else:
        for i in range(child_nodes.length):
            child_node = child_nodes.item(i)
            node_type = child_node.nodeType
            if node_type == Node.TEXT_NODE:
                fnode = child_node.nodeValue
                snode = fnode.strip()

                if len(snode) == 0:
                    continue

                if i > 0 and len(fnode.lstrip()) < len(fnode):
                    child = f" {snode}"

                elif i < (child_nodes.length - 1) and len(fnode.rstrip()) < len(fnode):
                    child = f"{snode} "

                else:
                    child = snode

                element.children.append(child)

            elif node_type == Node.ELEMENT_NODE:
                if (
                    child_node.tagName == "table:table-cell"
                    and "table:number-columns-repeated" in element.attributes
                ):
                    repeats = int(element.attributes["table:number-columns-repeated"])
                else:
                    repeats = 1

                for i in range(repeats):
                    element.children.append(parse_element(element_map, child_node))

            else:
                raise OdioException(f"Node type {node_type} not recognised.")

    return element.unpack()
